#!/usr/bin/python3
#
# Program to convert an Intel HEX file produced by SDCC into an MTX *.RUN file
#
import  sys
import  os

class MakeError (Exception):
    def __init__ (self, sMsg, nLine, sLine):
        self.sMsg = sMsg
        self.nLine = nLine
        self.sLine = sLine

    def __str__ (self):
        if ( self.sLine is not None ):
            return 'Error in line {:d}: {:s}\n    {:s}'.format (self.nLine, self.sMsg, self.sLine)
        else:
            return 'Error in line {:d}: {:s}'.format (self.nLine, self.sMsg)

class MakeRun:
    def __init__ (self):
        self.laddr = []
        self.data = {}
        
    def Load (self, sIn):
        print ('Loading file:  ' + sIn)
        with open (sIn, 'r') as fIn:
            nLine = 0
            for sLine in fIn:
                sLine = sLine.strip ()
                nLine += 1
                if ( not sLine.startswith (':') ):
                    raise MakeError ('Missing starting colon', nLine, sLine)
                bLine = bytes.fromhex (sLine[1:])
                if ( len (bLine) < 5 ):
                    raise MakeError ('Line too short', nLine, sLine)
                if ( len (bLine) != bLine[0] + 5 ):
                    raise MakeError ('Line length inconsistent with indicator', nLine, sLine)
                if ( sum (bLine) % 256 != 0 ):
                    raise MakeError ('Invalid checksum', nLine, sLine)
                if ( bLine[3] == 0 ):
                    addr = 256 * bLine[1] + bLine[2]
                    if ( addr in self.data ):
                        raise MakeError ('Duplicate address', nLine, sLine)
                    self.laddr.append (addr)
                    self.data[addr] = (nLine, bLine[4:-1])
                elif ( bLine[3] == 1 ):
                    break
                else:
                    raise MakeError ('Unsupported record type', nLine, sLine)
        print ('Records:       {:d}'.format (nLine))
                
    def Save (self, sOut):
        self.laddr.sort ()
        saddr = self.laddr[0]
        eaddr = self.laddr[-1]
        eaddr += len (self.data[eaddr][1])
        alen = eaddr - saddr
        with open (sOut, 'wb') as fOut:
            fOut.write (bytes ((saddr % 256, saddr // 256, alen % 256, alen // 256)))
            caddr = saddr
            npad = 0
            for addr in self.laddr:
                if ( addr < caddr ):
                    raise MakeError ('Records overlap', self.data[addr][0], '{:04X}'.format (addr))
                elif ( addr > caddr ):
                    fOut.write (bytes (addr - caddr))
                    npad += addr - caddr
                fOut.write (self.data[addr][1])
                caddr = addr + len (self.data[addr][1])
        print ('Saved file:    ' + sOut)
        print ('Start address: {:04X}h'.format (saddr))
        print ('End address:   {:04X}h'.format (eaddr))
        print ('Length:        {:04X}h'.format (alen))
        print ('Padding:       {:04X}h'.format (npad))

    def Process (self, arg):
        narg = len (arg)
        if (( narg == 2 ) or ( narg == 3 )):
            if ( os.path.exists (arg[1]) ):
                sIn = arg[1]
            else:
                sIn = arg[1] + '.ihx'
                if ( not os.path.exists (sIn) ):
                    sys.stderr.write ('Input file "{:s}" not found'.format (arg[1]))
                    sys.exit (2)
            if ( narg == 3 ):
                if ( os.path.isdir (arg[2]) ):
                    sOut = os.path.join (arg[2], os.path.splitext (os.path.basename (arg[1]))[0]
                                         + '.run')
                else:
                    sOut = arg[2]
            else:
                sOut = os.path.splitext (arg[1])[0] + '.run'
        else:
            sys.stderr.write ('Usage: {:s} <Intel hex file> [<Output file | folder>]'.format (arg[0]))
            sys.exit (1)
        try:
            self.Load (sIn)
            self.Save (sOut)
        except MakeError as e:
            sys.stderr.write (str (e))
            sys.exit (2)

MakeRun ().Process (sys.argv)

            
            
