// vdpaddr.c - Set a VDP table address

#include <mtxvdp.h>

#ifndef __SDCCCALL
#define __SDCCCALL  0
#endif

/**
    Set a VDP table address. This is a convenience routine for configuring
    VDP registers 2 - 6. It shifts the address to the correct alignment
    in the register, and ensures that the LSB are all zero. Note that these
    checks do not prevent a table from wrapping across the end of VRAM.

    Note: To make use of the undocumented enhancements of graphics mode 2
    use vdpreg() to set registers 3 & 4.

    table =     0 - Name table
    table =     1 - Colour table
    table =     2 - Pattern table
    table =     3 - Sprite attributes table
    table =     4 - Sprite pattern table
    table =     5 - Graphics 2 colour table (addr = 0x0000 or 0x2000)
    table =     6 - Graphics 2 pattern table (addr = 0x0000 or 0x2000)
    addr =      Address for table
    Return:     0 - Success
    Return:     1 - Invalid table number
    Return:     2 - Address > 16K
    Return:     3 - Invalid LSB

*/

unsigned char vdpaddr (unsigned char table, unsigned int addr) __naked
    {
    table; addr;
    __asm
#if __SDCCCALL != 1
        ld      hl, #2      ; Pointer to parameters
        add     hl, sp
        ld      a, (hl)     ; A = table
        inc     hl
        ld      e, (hl)     ; DE = addr
        inc     hl
        ld      d, (hl)
#endif
        cp      #7          ; Test for valid table number
        jr      c, 00001$
#if __SDCCCALL == 1
        ld      a, #1
#else
        ld      l, #1
#endif
        ret
00001$:
        ld      c, a        ; Save table number
        ld      a, d        ; Check high bits of address
        cp      #0x40
        jr      c, 00002$
#if __SDCCCALL == 1
        ld      a, #2
#else
        ld      l, #2
#endif
        ret
00002$:
        ld      b, #0       ; Offset into shift table
        ld      hl, #00020$
        add     hl, bc
        ld      a, (hl)
        cp      #0x80       ; Test for negative
        jr      c, 00004$
00003$:
        sla     e           ; Shift up DE
        rl      d
        inc     a           ; Count shifts
        jr      nz, 00003$
        jr      00006$
00004$:
        ld      b, a
00005$:
        srl     d           ; Shift down D
        jr      c, 00007$   ; Jump if a non-zero LSB
        djnz    00005$      ; Repeat for each shift
00006$:
        ld      a, e        ; Check low byte is zero
        and     a
        jr      z, 00008$
00007$:
#if __SDCCCALL == 1
        ld      a, #3
#else
        ld      l, #3
#endif
        ret
00008$:
        ld      a, c        ; Test for G2 values
        cp      #5
        jr      c, 00009$
        inc     hl          ; Get mask bits
        inc     hl
        ld      l, (hl)
        ld      a, d        ; Test for bits set
        and     l
        jr      nz, 00007$
        ld      a, d        ; Set the lsb to one
        or      l
        ld      d, a
        ld      a, c        ; Adjust register number
        sub     #4
        ld      c, a
00009$:
        call    getiff      ; Get interrupt status
        push    af
        di
        in      a,(0x02)    ; Read status register
        ld      a, d        ; Output aligned address
        out     (0x02), a
        ld      a, c        ; Select register
        add     a, #2
        or      #0x80
        out     (0x02), a
        pop     af          ; Optionally enable interrupts
        jr      c, 00010$
        ei
00010$:
        xor     a           ; Return zero
#if __SDCCCALL != 1
        ld      l, a
#endif
        ret
00020$:
        .db     2, -2, 3, -1, 3, -2, 3      ; Amount to shift address
        .db     0x7F, 0x03                  ; G2 mask bits
    __endasm;
    }
